{ FFT_UNIT - integer-type Fast Fourier Transform routines }
{ for Borland Pascal }

{ (c) Copyright 1997 by Alexey Monastyrenko }

{ If you use this routines in your program, please remember reference to me }
{ ᫨  ᯮ  楤  ᢮ ணࠬ,   㯮  }

unit FFT_Unit;

interface

type TComplex=record Re,IM:Integer end;
const N=256;
const Log2N=8;

type TComplexArray=array[0..N-1] of TComplex;

procedure FFT(var InBuf,OutBuf:TComplexArray);

procedure FFTEnergy(var Buf:TComplexArray); {Re, Im - 2 }

implementation

var FFTExp:TComplexArray;

var BitReverse:array[0..N-1] of Integer;

const FixedBase:Integer=16384;

function FMul(X, Y: Integer): Integer;
inline($5A/     {pop  dx}
       $58/     {pop  ax}
       $F7/$EA/ {imul dx}
       $0F/$AC/$D0/$0E   {shrd ax,dx,0E}
       );

procedure InitExp;
var I:Integer;
begin
  for I:=0 to N-1 do begin
    FFTExp[I].Re:= Round(Cos(I*2*PI/N)*FixedBase);
    FFTExp[I].Im:=-Round(Sin(I*2*PI/N)*FixedBase);
  end;
end;

procedure InitBitReverse;
var I,J:Integer;
begin
  for I:=0 to N-1 do BitReverse[I]:=0;
  for J:=0 to Log2N-1 do begin
    for I:=0 to N-1 do
     BitReverse[I]:=BitReverse[I]*2+((I shr J) and 1);
  end;
end;

procedure FFT2(var Buf:TComplexArray);
var I,J,K,Step,Step2:Integer;
    Tmp:TComplex;
begin
  Step:=1; Step2:=N div 2;
  while Step<>N do begin
    for J:=0 to Step2 - 1 do begin
      for K:=0 to Step - 1 do begin
        I:=J*Step*2 + K;
        Tmp.Re := FMul(Buf[I+Step].Re,FFTExp[K*Step2].Re)
                 -FMul(Buf[I+Step].Im,FFTExp[K*Step2].Im);
        Tmp.Im := FMul(Buf[I+Step].Re,FFTExp[K*Step2].Im)
                 +FMul(Buf[I+Step].Im,FFTExp[K*Step2].Re);
        Buf[I+Step].Re:=(Buf[I].Re-Tmp.Re);
        Buf[I+Step].Im:=(Buf[I].Im-Tmp.Im);
        Buf[I].Re:=(Buf[I].Re+Tmp.Re);
        Buf[I].Im:=(Buf[I].Im+Tmp.Im);
      end;
    end;
    Step:=Step*2; Step2:=Step2 div 2;
  end;
end;

procedure FFT(var InBuf,OutBuf:TComplexArray); {Complex number FFT}
var I:Integer;
begin
  for I:=0 to N-1 do OutBuf[I]:=InBuf[BitReverse[I]];
  FFT2(OutBuf);
end;

function FEnergy(X, Y: Integer): Word; {(X^2+Y^2)/N^2}
inline(
       $58/     {pop  ax}
       $C1/$F8/<Log2N-2/ {Sar ax, log2N-2}
       $F7/$E8/ {imul ax}
       $89/$C3/ {mov  bx,ax}
       $89/$D1/ {mov  cx,dx}
       $58/     {pop  ax}
       $C1/$F8/<Log2N-2/ {Sar ax, log2N-2}
       $F7/$E8/ {imul ax}
       $01/$D8/ {add  ax,bx}
       $11/$CA  {adc  dx,cx}
);

procedure FFTEnergy(var Buf:TComplexArray); {Re, Im - 2 channels}
{Evaluates harmonics ampl. squares at 2 channel}
{   ମ  2- }
var I:Integer;
    Buf2:TComplexArray;
begin
  FFT(Buf,Buf2);
  Buf[0].Re:=FEnergy(Buf2[0].Re,0);
  Buf[0].IM:=FEnergy(Buf2[0].IM,0);
  for I:=1 to N div 2 - 1 do begin
    Buf[I].Re:=FEnergy(Buf2[I].Re+Buf2[N-I].Re,Buf2[N-I].IM-Buf2[I].IM);
    Buf[I].IM:=FEnergy(Buf2[I].IM+Buf2[N-I].IM,Buf2[I].Re-Buf2[N-I].Re);
  end;
end;

begin
  InitBitReverse;
  InitExp;
end.